# Standard imports
import io
import os
import sys

# Local imports
from oblivion.constants import (
    callback,
    HASH_LENGTH,
)
from oblivion.entities import (
    RSAPublicKey,
)
from oblivion.schemes import (
    rsa_encryption_scheme_with_oaep_padding,
)


def put_subparser(subparsers):
    parser = subparsers.add_parser(
        'encrypt',
        help='Encrypt a file read from standard input (/dev/stdin)',
    )
    parser.add_argument(
        '-k', '--key-name',
        help='file name of the key, defaults to "rsa"',
        required=False,
        default='rsa',
    )
    parser.set_defaults(subparser_handler=handler)


def handler(args):
    key = RSAPublicKey(modulus=0, exponent=0)

    path: str = \
        os.path.abspath(
            os.path.join(
                os.getcwd(),
                f'{args.key_name}.{key.kind}.{key.extension}'))

    callback(f'reading {key.kind} key from: {path}')
    if not os.path.exists(path):
        callback(f'error: no such key')
        sys.exit(1)

    key.load_from_file(path)

    # We can encrypt at much this data octets per round
    block_size = key.modulus_octets - 2 * (HASH_LENGTH + 1)
    # Adjust the last octet
    block_size -= 1

    with io.BufferedReader(sys.stdin.buffer) as buffer:
        block = buffer.read(block_size)
        while block:
            encrypted_block = rsa_encryption_scheme_with_oaep_padding(
                public_key=key,
                message=block,
                label=b'oblivion',
            )
            print(encrypted_block.hex())
            block = buffer.read(block_size)



# Standard imports
import json
from typing import Dict, Union

# Local imports
from oblivion.constants import (
    RSAModulus,
    RSAExponent,
)


class RSAKey():
    """Class to represent a RSA Key."""
    # pylint: disable=too-many-instance-attributes

    kind: str = 'generic'
    extension: str = 'json'

    def __init__(self, *,
                 modulus: RSAModulus,
                 exponent: RSAExponent):
        """Class constructor."""
        self._modulus: RSAModulus = 0
        self.modulus_octets: int = 0
        self.modulus_bits: int = 0
        self.modulus: RSAModulus = modulus
        self._exponent: RSAExponent = 0
        self.exponent_bits: int = 0
        self.exponent: RSAExponent = exponent

    @property
    def exponent(self) -> RSAExponent:
        return self._exponent

    @exponent.setter
    def exponent(self, exponent: RSAExponent):
        self._exponent = exponent
        self.exponent_bits = exponent.bit_length()

    @property
    def modulus(self) -> RSAModulus:
        return self._modulus

    @modulus.setter
    def modulus(self, modulus: RSAModulus):
        self._modulus = modulus
        self.modulus_bits = modulus.bit_length()
        self.modulus_octets = (
            self.modulus_bits // 8 + (self.modulus_bits % 8 != 0))

    def save_to_file(self, path: str):
        """Save the Key to a location in the file system."""
        with open(path, 'w') as target_handle:
            data: Dict[str, Union[int, str]] = {
                'kind': self.kind,
                'modulus': self.modulus,
                'exponent': self.exponent,
                'modulus_bits': self.modulus_bits,
                'exponent_bits': self.exponent_bits,
            }
            target_handle.write(json.dumps(data, indent=4))
            target_handle.write('\n')

    def load_from_file(self, path: str):
        """Load the Key from a location in the file system."""
        with open(path, 'r') as target_handle:
            data: Dict[str, Union[int, str]] = json.loads(target_handle.read())
            self.kind = str(data['kind'])
            self.modulus = int(data['modulus'])
            self.exponent = int(data['exponent'])


class RSAPublicKey(RSAKey):
    """Class to represent a RSA Public Key."""

    kind: str = 'public'


class RSAPrivateKey(RSAKey):
    """Class to represent a RSA Private Key."""

    kind: str = 'private'



class RSAException(Exception):
    """Base exceptions for this package."""


class RSAEncryptionException(RSAException):
    """An error ocurred while encrypting."""


class RSADecryptionException(RSAException):
    """An error ocurred while decrypting."""



# Standard imports
from typing import Tuple

# Local imports
from oblivion.constants import (
    RSAMsgRepresentative,
    RSACyphertextRepresentative,
)
from oblivion.entities import (
    RSAPublicKey,
    RSAPrivateKey,
)


def octet_string_to_integer_primitive(octet_string: bytes) -> Tuple[int, int]:
    """Convert an octet string to a nonnegative integer."""
    # https://tools.ietf.org/html/rfc8017#section-4.2
    integer = \
        sum(value * (256 ** index)
            for index, value in enumerate(reversed(octet_string)))
    return len(octet_string), integer


def integer_to_octet_string_primitive(length: int, integer: int) -> bytes:
    """Convert a nonnegative integer to an octet string."""
    # https://tools.ietf.org/html/rfc8017#section-4.1
    if integer >= 256 ** length:
        raise ValueError(f"{integer} >= 256 ** {length}")
    index = 0
    digits = [0] * length
    while integer:
        digits[index] = integer % 256
        integer //= 256
        index += 1
    return bytes(reversed(digits))


def rsa_encryption_primitive(public_key: RSAPublicKey,
                             message: RSAMsgRepresentative,
                             ) -> RSACyphertextRepresentative:
    # https://tools.ietf.org/html/rfc8017#section-5.1.1
    if not 0 <= message <= public_key.modulus:
        raise ValueError(f'not 0 <= {message} <= {public_key.modulus}')
    return pow(message, public_key.exponent, public_key.modulus)


def rsa_decryption_primitive(private_key: RSAPrivateKey,
                             ciphertext: RSACyphertextRepresentative,
                             ) -> RSACyphertextRepresentative:
    # https://tools.ietf.org/html/rfc8017#section-5.1.2
    if not 0 <= ciphertext <= private_key.modulus:
        raise ValueError(f'not 0 <= {ciphertext} <= {private_key.modulus}')
    return pow(ciphertext, private_key.exponent, private_key.modulus)



# Standard imports
import os

# Local imports
from oblivion.constants import (
    callback,
    MINIMUM_MODULUS_BIT_SIZE,
)
from oblivion.functions import (
    rsa_generate_keys,
)


def put_subparser(subparsers):
    parser = subparsers.add_parser(
        'gen',
        help='Generates a new RSA Key Pair',
    )
    parser.add_argument(
        '-n', '--name',
        help='file name where the key will be stored, defaults to "rsa"',
        required=False,
        default='rsa',
    )
    parser.add_argument(
        'bits',
        help='minimum bits on the RSA modulus, (4196, 16384, 65536, etc)',
        type=int,
    )
    parser.set_defaults(subparser_handler=handler)


def handler(args):
    callback('generating a new secure RSA key pair')
    if args.bits < MINIMUM_MODULUS_BIT_SIZE:
        callback(f'minimum bit size is {MINIMUM_MODULUS_BIT_SIZE}, adjusting')
        args.bits = MINIMUM_MODULUS_BIT_SIZE
    public_key, private_key = rsa_generate_keys(args.bits)

    for key in (public_key, private_key):
        path: str = \
            os.path.abspath(
                os.path.join(
                    os.getcwd(),
                    f'{args.name}.{key.kind}.{key.extension}'))

        callback(f'saving {key.kind} key to: {path}')

        key.save_to_file(path)



# Local imports
from oblivion.constants import (
    RSAMsg,
    RSALabel,
    RSACyphertext,
    HASH_LENGTH,
)
from oblivion.entities import (
    RSAPublicKey,
    RSAPrivateKey,
)
from oblivion.exceptions import (
    RSAEncryptionException,
    RSADecryptionException,
)
from oblivion.primitives import (
    octet_string_to_integer_primitive,
    integer_to_octet_string_primitive,
    rsa_encryption_primitive,
    rsa_decryption_primitive,
)
from oblivion.functions import (
    hash_func,
    random_bytes,
    bytes_xor,
    mask_generation_function,
)


def rsa_encryption_scheme_with_oaep_padding(public_key: RSAPublicKey,
                                            message: RSAMsg,
                                            label: RSALabel = b'',
                                            ) -> bytes:
    """Perform RSAES-OAEP-ENCRYPT as per RFC-8017."""
    # https://tools.ietf.org/html/rfc8017#section-7.1.1
    # pylint: disable=too-many-locals
    message_length = len(message)
    if message_length > (public_key.modulus_octets - 2 * (HASH_LENGTH + 1)):
        raise RSAEncryptionException()

    label_hash = hash_func(label)
    padding_string = b'\x00' * (
        public_key.modulus_octets
        - message_length
        - 2 * (HASH_LENGTH + 1))
    data_block = label_hash + padding_string + b'\x01' + message
    seed = random_bytes(HASH_LENGTH)
    data_block_mask = mask_generation_function(
        seed, public_key.modulus_octets - HASH_LENGTH - 1)
    masked_data_block = bytes_xor(data_block, data_block_mask)
    seed_mask = mask_generation_function(masked_data_block, HASH_LENGTH)
    masked_seed = bytes_xor(seed, seed_mask)
    encoded_message = b'\x00' + masked_seed + masked_data_block
    _, encoded_message_representative = \
        octet_string_to_integer_primitive(encoded_message)
    ciphertext_representative = rsa_encryption_primitive(
        public_key=public_key,
        message=encoded_message_representative)
    ciphertext = integer_to_octet_string_primitive(
        length=public_key.modulus_octets,
        integer=ciphertext_representative)
    return ciphertext


def rsa_decryption_scheme_with_oaep_padding(
        private_key: RSAPrivateKey,
        ciphertext: RSACyphertext,
        label: RSALabel = b'') -> bytes:
    """Perform RSAES-OAEP-DECRYPT as per RFC-8017."""
    # https://tools.ietf.org/html/rfc8017#section-7.1.2
    # pylint: disable=too-many-locals
    ciphertext_length = len(ciphertext)

    if ciphertext_length != private_key.modulus_octets \
            or private_key.modulus_octets < 2 * (HASH_LENGTH + 1):
        raise RSADecryptionException()

    _, ciphertext_representative = \
        octet_string_to_integer_primitive(ciphertext)
    message_representative = rsa_decryption_primitive(
        private_key=private_key,
        ciphertext=ciphertext_representative)
    encoded_message = integer_to_octet_string_primitive(
        length=private_key.modulus_octets,
        integer=message_representative)
    label_hash = hash_func(label)
    masked_byte = encoded_message[0]
    masked_seed = encoded_message[1:1 + HASH_LENGTH]
    masked_data_block = encoded_message[1 + HASH_LENGTH:]
    seed_mask = mask_generation_function(masked_data_block, HASH_LENGTH)
    seed = bytes_xor(masked_seed, seed_mask)
    data_block_mask = mask_generation_function(
        mgf_seed=seed,
        mask_length=private_key.modulus_octets - HASH_LENGTH - 1)
    data_block = bytes_xor(masked_data_block, data_block_mask)
    expected_label_hash = data_block[:HASH_LENGTH]
    index = data_block[HASH_LENGTH:].find(b'\x01')
    if index == -1 \
            or masked_byte != 0x00 \
            or label_hash != expected_label_hash:
        raise RSADecryptionException()
    message = data_block[HASH_LENGTH + index + 1:]
    return message



# Standard imports
import argparse

# Local imports
from oblivion.cli import (
    gen,
    encrypt,
    decrypt,
)


def cli():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        dest='subparser_name')

    gen.put_subparser(subparsers)
    encrypt.put_subparser(subparsers)
    decrypt.put_subparser(subparsers)

    args = parser.parse_args()

    if args.subparser_name:
        args.subparser_handler(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    cli()



# Standard imports
import io
import os
import sys

# Local imports
from oblivion.constants import (
    callback,
)
from oblivion.entities import (
    RSAPrivateKey,
)
from oblivion.schemes import (
    rsa_decryption_scheme_with_oaep_padding,
)


def put_subparser(subparsers):
    parser = subparsers.add_parser(
        'decrypt',
        help='Decrypt a file read from standard input (/dev/stdin)',
    )
    parser.add_argument(
        '-k', '--key-name',
        help='file name of the key, defaults to "rsa"',
        required=False,
        default='rsa',
    )
    parser.set_defaults(subparser_handler=handler)


def handler(args):
    key = RSAPrivateKey(modulus=0, exponent=0)

    path: str = \
        os.path.abspath(
            os.path.join(
                os.getcwd(),
                f'{args.key_name}.{key.kind}.{key.extension}'))

    callback(f'reading {key.kind} key from: {path}')
    if not os.path.exists(path):
        callback(f'error: no such key')
        sys.exit(1)

    key.load_from_file(path)

    with io.BufferedReader(sys.stdin.buffer, 1024) as buffer, \
            io.BufferedWriter(sys.stdout.buffer, 1024) as out_buffer:
        block = bytes.fromhex(buffer.readline().decode()[0:-1])
        while block:
            decrypted_block = rsa_decryption_scheme_with_oaep_padding(
                private_key=key,
                ciphertext=block,
                label=b'oblivion',
            )
            out_buffer.write(decrypted_block)
            block = bytes.fromhex(buffer.readline().decode()[0:-1])



# Standard imports
import math
import secrets
from typing import Tuple

# Local imports
from oblivion.constants import (
    HashBase,
    HASH_LENGTH,

    callback,
)
from oblivion.entities import (
    RSAPublicKey,
    RSAPrivateKey,
)
from oblivion.primitives import (
    integer_to_octet_string_primitive,
)


def hash_func(msg: bytes) -> bytes:
    """Return the digest of msg."""
    return HashBase(msg).digest()


def random_bytes(size: int) -> bytes:
    """Return a safe random octet string."""
    return secrets.token_bytes(size)


def bytes_xor(bytes_a: bytes, bytes_b: bytes) -> bytes:
    """Return the XOR between A and B."""
    if len(bytes_a) != len(bytes_b):
        raise ValueError('Expected equal length octet strings')
    return bytes(a ^ b for a, b in zip(bytes_a, bytes_b))


def jacobi(numerator: int, denominator: int) -> int:
    """Compute the Jacobi Symbol."""
    if denominator <= 0 or denominator & 1 == 0:
        raise ValueError('Jacobi parameters are out of function domain')
    j_symbol: int = 1
    numerator %= denominator
    while numerator:
        while numerator & 1 == 0:
            numerator = numerator // 2
            if denominator % 8 in (3, 5):
                j_symbol *= -1
        numerator, denominator = denominator, numerator
        if numerator % 4 == denominator % 4 == 3:
            j_symbol *= -1
        numerator %= denominator
    if denominator == 1:
        return j_symbol
    return 0


def is_prime(number, rounds: int = 16) -> bool:
    """Primality test via Solovay-Strassen algorithm."""
    helper: int = (number - 1) // 2
    for _ in range(rounds):
        witness: int = 1 + secrets.randbelow(number - 1)
        if math.gcd(witness, number) != 1:
            return False
        if jacobi(witness, number) % number != pow(witness, helper, number):
            return False
    return True


def generate_prime(bits: int) -> int:
    """Return a prime number of the specified bits."""
    callback('generating prime')
    while True:
        prime = secrets.randbits(bits + 1)
        prime += (1 - prime & 1)
        if (prime).bit_length() < bits:
            continue
        if is_prime(prime):
            return prime


def generate_special_prime(bits: int, depth: int = 2) -> int:
    """
    Return a prime 'p' where 'p - 1' has large prime factors.

    To guarantee that 'p - 1' has a large prime factor we'll generate a random
    base prime number 'u', and then find the first prime 'p' in the sequence:

        i * u + 1, where i = 2, 4, 6, ...

    With this, we ensure that the only prime factor of 'p - 1' is 'u', which is
    large.

    To provide extra security, 'u - 1' will also contain large prime factors,
    and recursively the base prime used to generate 'u', up to the provided
    depth level.
    """
    callback('generating base prime')

    counter: int = 0
    if depth > 1:
        base_prime = generate_special_prime(bits, depth=depth - 1)
    else:
        base_prime = generate_prime(bits)

    callback('incrementing base prime')
    while True:
        counter += 2
        prime = counter * base_prime + 1
        if is_prime(prime):
            return prime


def extended_euclidian_gcd(left: int, right: int) -> Tuple[int, int, int]:
    """Solve left * x + right * y = gcd(left, right) and return r, x, y."""
    new_x, new_y, new_r = 0, 1, right
    old_x, old_y, old_r = 1, 0, left
    while new_r != 0:
        quotient = old_r // new_r
        old_r, new_r = new_r, old_r - quotient * new_r
        old_x, new_x = new_x, old_x - quotient * new_x
        old_y, new_y = new_y, old_y - quotient * new_y
    return old_r, old_x, old_y


def modular_multiplicative_inverse(number: int, modulo: int) -> int:
    """Compute the multiplicative inverse of number under given modulo."""
    # Assumes number and modulo are co-prime
    return extended_euclidian_gcd(number, modulo)[1] % modulo


def rsa_generate_keys(bits: int) -> Tuple[RSAPublicKey, RSAPrivateKey]:
    """Generate public and private keys with (n, d) of the specified bits."""
    public = RSAPublicKey(modulus=0, exponent=0)
    private = RSAPrivateKey(modulus=0, exponent=0)
    reduced_totient_divisor_min_bits: int = bits // 2 ** 4
    while True:
        callback('generating prime number: p')
        # p and q should differ in length by a few digits
        # both (p - 1) and (q - 1) should contain large prime factors
        prime_p = generate_special_prime(bits // 2 - 1)
        callback('generating prime number: q')
        prime_q = generate_special_prime(bits // 2 + 1)
        callback('computing modulus: n')
        public.modulus = prime_p * prime_q
        private.modulus = public.modulus
        callback('computing health check')
        # gcd(p - 1, q - 1) should be small
        reduced_totient_divisor: int = math.gcd(prime_p - 1, prime_q - 1)
        if public.modulus_bits >= bits:
            if reduced_totient_divisor_min_bits:
                if (reduced_totient_divisor.bit_length()
                        < reduced_totient_divisor_min_bits):
                    break
            else:
                break
        callback('health check failed, retrying')
    callback('computing reduced_totient of: n')
    reduced_totient = (prime_p - 1) * (prime_q - 1) // reduced_totient_divisor
    while True:
        callback('computing private exponent: d')
        private.exponent = generate_prime(bits)
        # It's guaranteed that private.exponent < reduced_totient (FIPS 186-4)
        callback('computing public exponent: e')
        public.exponent = \
            modular_multiplicative_inverse(private.exponent, reduced_totient)
        callback('computing health check')
        # Ensure that any message (except for 0 and 1), will be reduced
        #   after performing encryption, this is, M ** e (mod n) will
        #   'wrap around'
        if public.modulus_bits < public.exponent < reduced_totient \
                and math.gcd(public.exponent, reduced_totient) == 1:
            break
        callback('health check failed, retrying')
    return public, private


def mask_generation_function(mgf_seed: bytes, mask_length: int) -> bytes:
    """Mask generation function based on a hash function."""
    # https://tools.ietf.org/html/rfc8017#appendix-B.2.1
    if mask_length > (2 ** 32) * HASH_LENGTH:
        raise ValueError(f'{mask_length} > {(2 ** 32) * HASH_LENGTH}')

    range_length = 1 + mask_length // HASH_LENGTH
    if mask_length % HASH_LENGTH != 0:
        range_length -= 1

    mgf_output = bytes()
    for counter in range(1 + range_length):
        mgf_output += hash_func(
            mgf_seed
            + integer_to_octet_string_primitive(
                length=4,
                integer=counter))
    return mgf_output[:mask_length]



"""Setup package."""

import distutils
from datetime import datetime


def get_minor_version() -> int:
    """Number of seconds since the beginning of the month."""
    utc_now = \
        datetime.utcnow()
    utc_beginning_of_month = \
        datetime.utcnow().replace(day=1, minute=0, second=0)
    return int((utc_now - utc_beginning_of_month).total_seconds())


def get_version():
    return datetime.utcnow().strftime(f'%Y.%m.{get_minor_version()}')


distutils.core.setup(
    name='oblivion',
    version=get_version(),
    packages=[
        'oblivion',
        'oblivion.cli'
    ],

    install_requires=[
    ],
    entry_points={
        'console_scripts': [
            'oblivion=oblivion.cli:cli',
        ],
    },
)



# Standard imports
import sys
import hashlib
import textwrap
from typing import Any
from inspect import getouterframes, currentframe

# Types
RSAModulus = int
RSAExponent = int
RSAPublicExponent = int
RSAPrivateExponent = int
RSAMsg = bytes
RSALabel = bytes
RSACyphertext = bytes
RSAMsgRepresentative = int
RSACyphertextRepresentative = int

# Crypto constants
HashBase = hashlib.sha3_512
HASH_LENGTH = HashBase().digest_size
MINIMUM_MODULUS_BIT_SIZE = 2 * (HASH_LENGTH + 1) * 8


def callback(msg: str) -> Any:
    """Callback function to notify progress."""
    current_depth: int = len(getouterframes(currentframe()))
    print(textwrap.indent(msg, prefix='-' * current_depth), file=sys.stderr)
